import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.semanticweb.owlapi.*;
import org.semanticweb.owlapi.model.*;
import org.semanticweb.owlapi.reasoner.ConsoleProgressMonitor;
import org.semanticweb.owlapi.reasoner.InferenceType;
import org.semanticweb.owlapi.reasoner.NodeSet;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerConfiguration;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;
import org.semanticweb.owlapi.reasoner.SimpleConfiguration;
import org.semanticweb.owlapi.reasoner.structural.StructuralReasonerFactory;
import org.semanticweb.owlapi.util.OWLOntologyWalker;
import org.semanticweb.owlapi.util.OWLOntologyWalkerVisitor;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.HermiT.Reasoner;
import org.semanticweb.owlapi.reasoner.Node;
import org.semanticweb.owlapi.reasoner.NodeSet;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import uk.ac.manchester.cs.owl.owlapi.OWLDisjointClassesAxiomImpl;


public class ImportOwl {
   
   //Sets up the variables needed.
   static OWLReasonerFactory reasonerFactory = new StructuralReasonerFactory();
   static OWLOntologyManager m = OWLManager.createOWLOntologyManager();
   static OWLOntology o;
   static OWLDataFactory df = OWLManager.getOWLDataFactory();
   static PrintWriter axiomwriter;
   private static String axiomAppendInfo ="";
   private static String specificClassPretext = "abstract OntologySpecificClasses = Cat **\n{\ncat\nInd; Prop;\n\nfun";
   private static String specificClassAppendInfo = " : Prop;";
		
      
   public ImportOwl() {	   
   }
	//Converts from the axioms to functionalSyntax
   //Does this by iterating over each axiom and cleaning up text related to each.
   public static void toFunctionalSyntax(String ontologyName, Boolean displayOutput) throws UnsupportedEncodingException  {
      try {
      //load ontology using ontologyName
         o = m.loadOntologyFromOntologyDocument(IRI.create(new File("Ont/"+ontologyName)));
      
      } 
      catch (OWLOntologyCreationException e) {
         System.out.println("ERROR! Ontology '"+ontologyName+ " not found in Ont/ directory.");
         e.printStackTrace();
      }
      try {
    	  //outputfile
         axiomwriter = new PrintWriter("functionalSyntax.txt", "UTF-8");
      } 	   
      catch (FileNotFoundException e) {
         System.out.println("ERROR! Ontology file not found!");
         e.printStackTrace();
      }
      printAll(o.getAxioms(), displayOutput);
   
      axiomwriter.close();
      System.out.println("Total axioms: "+o.getAxioms().size());
   }
   
   private static void printAll(Set axioms, Boolean displayOutput) {
      Iterator owlaxioms = axioms.iterator();
      int i=0;
      while (owlaxioms.hasNext()) { 
         OWLAxiom axiom = (OWLAxiom) owlaxioms.next();
         AxiomType<?> type = axiom.getAxiomType();
         if (checkIfPrint(type) == false)  {
            continue;
         }
         i++;
         String axiomString = "";
         if (type == AxiomType.DISJOINT_CLASSES ) {
            Set<OWLDisjointClassesAxiom> disjointAxioms =  ((OWLDisjointClassesAxiom) axiom).asPairwiseAxioms();
            for (OWLDisjointClassesAxiom ax: disjointAxioms){
               axiomString = clean(ax.toString());
               axiomString = stripDashes(axiomString);
               axiomString = preOverload(axiomString);
            }  
         }		
         else {
            axiomString = clean(axiom.toString());
            axiomString = stripDashes(axiomString);
            axiomString = preOverload(axiomString);
         }
         if (displayOutput) {
            System.out.println(axiomString);
         }
         axiomwriter.println(axiomString+axiomAppendInfo);
      }
      System.out.println("Axioms processesed: "+ i);
   }
	   
   //Getting around being unable to overload in GF.
   private static String preOverload(String axiomString) {
      String outS = axiomString;
      int count=0;
      String[] axiomArr = new String[] {"ObjectUnionOf", "ObjectIntersectionOf", "ObjectOneOf", "DifferentIndividuals", "DisjointUnion"};
      for(String ax: axiomArr) {
         if (axiomString.contains(ax)) {
            count=0;
            int start=axiomString.indexOf(ax);
            int end=0;
            int open = 0;
            String newString = axiomString.substring(start);
            for (int i = 0; i < newString.length(); i++) {
               char currentChar = newString.charAt(i);
               if (currentChar == '(') {
                  open++;
               }
               else if(currentChar == ')') {
                  open--;
                  if(open==-1) {
                     break;
                  }
                  if(open==0) {
                     count++;
                  }
               }
               else if((open==0) && (currentChar == ' ')){
                  if (((newString.charAt(i-1) == ')') ||  (newString.charAt(i+1) == '('))) {
                  }
                  else{
                     count++;
                  }
               }
            }
            outS = outS.replaceAll(ax, ax+count);
         }
      }
      return outS;
   }
   //Simple method to strip the dashes from an axiom string.
   private static String stripDashes(String axiomString) {
      if (axiomString.indexOf("-")==-1) {
      	//Decision to capitilise if there aren't dashes. This is to avoid having props which are reserved words in GF. (eg cat)
         return axiomString.substring(0,1).toUpperCase() + axiomString.substring(1);
      }
      String outS ="";
      String[] splitString= axiomString.split("-");
      for(int i=0;i<splitString.length;i++){
         String current =splitString[i];
         current = current.substring(0,1).toUpperCase() + current.substring(1);
         outS+=current;
      }
      return outS;
   		   
   }
   //clean method which turns the axiomstring into GF readable functional syntax
   public static String clean(String axiomString){ 
      axiomString = preProcessAxiomString(axiomString);
      String[] splitString = axiomString.split(" ");
      String outS="";
      
      //iterate over the string by spaces
      for (String s: splitString) {
         //If there are opening brackets, propagate them backwards for the GF syntax.
         if (s.indexOf("(") >-1) {
            String newString="";
            int lastIndex =s.lastIndexOf("(");
            newString = "("+s.substring(0, lastIndex)+ s.substring(lastIndex,s.length()).replace("(", " ");
            s=newString;
         }
         if(s.length() <3) {
            outS+=s;
            continue;
         }

         //Deal with the two major types of axiomstrings slightly differently depending on the #.
         else if (s.contains("<") && s.contains(">") && s.contains("#")) {
            int startIndex = s.indexOf("<");
            int hashIndex = s.indexOf("#");
            int endIndex = s.indexOf(">");
            String substring = s.substring(startIndex, endIndex+1);
            String value =  s.substring(hashIndex+1, endIndex);
            value = value.substring(0,1).toUpperCase() + value.substring(1);
            String newString = s.replace(substring,value);
            outS+= " " +newString;
         }
         else if (s.contains("<") && s.contains(">")) {
            int startIndex = s.indexOf("<");
            int hashIndex = s.lastIndexOf("/");
            int endIndex = s.indexOf(">");
            String substring = s.substring(startIndex, endIndex+1);
            String value =  s.substring(hashIndex+1, endIndex);
            String newString = s.replace(substring,value);
            outS+= " " +newString;
         
         	   
         }
         else {
            outS+=" "+s;
         }
      }
      return outS;
   
   }
   private static String preProcessAxiomString(String axiomString) {
      if(axiomString.contains("owl:")) {
         axiomString = axiomString.replace("owl:", "");
      
      }
      if(axiomString.contains("\"")) {
         if((axiomString.indexOf("\"") != (axiomString.lastIndexOf("\"")))) {
            axiomString= axiomString.replaceAll("\"", "");
         
         }
      }
      
      //a couple of hardcoded boundary cases
      if(axiomString.contains("^^xsd:integer")) {
         axiomString = axiomString.replace("^^xsd:integer", "");
      }
      if(axiomString.contains("^^xsd:positiveInteger")) {
         axiomString = axiomString.replace("^^xsd:positiveInteger", "");
         axiomString = axiomString.replace("\"", " "); 
      }
      if(axiomString.contains("^^xsd:boolean")) {
         axiomString = axiomString.replace("^^xsd:boolean", "");
         axiomString = axiomString.replace("\"", " "); 
      }
      if(axiomString.contains("xsd:")) {
         axiomString = axiomString.replace("xsd:", "");
      }
   
      return axiomString;
   }
   
   //Courtesy of: http://rosettacode.org/wiki/Count_occurrences_of_a_substring
   //Counts instances of substring in a string. Very cool solution which I found to the problem online. Chose to use it because its kind of beautiful.
   public static int countSubstring(String subStr, String str){
      return (str.length() - str.replace(subStr, "").length()) / subStr.length();
   }
   //Boolean method to return whether a axiom is ignored (return false) or not (return true)
   //We don't deal with annotations and declarations. This is a method which is easy to edit in the future if this needs to change.
   
   public static Boolean checkIfPrint(AxiomType<?> type) {
      if (type == AxiomType.ANNOTATION_ASSERTION) {
         return false;
      }
      else if (type == AxiomType.DECLARATION){
         return false;
      }
      else {
         return true;
      }
   }
}