import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Scanner;
import java.util.Set;
import java.util.Arrays;


public class ProcessFile {
   private static String gf_path = "C:\\Users\\Ion\\AppData\\Roaming\\cabal\\gf-3.7";
	//Extract the names from the text which need to be translated, don't include known axioms.
	//Return these as a set<String> to be used in the word lookup.
   public static Set<String> extractClasses(String filename) {
      Set<String> uniqueClasses = new HashSet<String>();
   	//Hardcoded list of axioms OWL 2 DL axioms which should be excluded, such as SubClass, ObjectSomeValuesFrom etc.
      Set<String> knownAxioms = new HashSet<String>(Arrays.asList( new String[] {
         	"ClassAssertion", "ObjectSomeValuesFrom", "ObjectPropertyRange", "ObjectExactCardinality", 
                "ObjectPropertyDomain", "InverseObjectProperties", "ObjectMaxCardinality", 
                "ObjectPropertyAssertion", "DisjointClasses", "ObjectAllValuesFrom", "SubClassOf", "Thing", 
                "EquivalentClasses", "SubObjectPropertyOf", "ObjectIntersectionOf", "1", "HasPart", "IsProperPartOf",
                "ReflexiveObjectProperty", "ObjectUnionOf", "IrreflexiveObjectProperty", "TransitiveObjectProperty",
                "ObjectComplementOf", "ObjectOneOf", "DataPropertyRange", "DataPropertyDomain", "SymmetricObjectProperty",
                "DifferentIndividuals", "ObjectHasValue", "ObjectMinCardinality", "ObjectMaxCardinality", "FunctionalObjectProperty", 
                "InverseFunctionalObjectProperty",  "DisjointUnion", "InverseOf", "DataSomeValuesFrom", "SymmetricObjectProperty", 
                "DataPropertyAssertion"
         }));
   	
      try (BufferedReader br = new BufferedReader(new FileReader(filename))) {
         String line;
         while ((line = br.readLine()) != null) {
            line=line.replace("(", " ");
            line=line.replace(")", " ");
            line=line.trim();
          	//System.out.println(line);
            String[] lineArr = line.split(" ");
            for(String word: lineArr) {
            	//Ignore single length words
               if(word.length()<=1) {
                  continue;
               }
               if(isNumeric(word)) {
                  continue;
               }
            	//It would be neater to do this as one if, but its faster to do it as two.
               if(!(knownAxioms.contains(word))){
                  if(!(word.contains("ObjectUnionOf") || (word.contains("ObjectIntersectionOf")) || 
                  		(word.contains("ObjectOneOf")) || (word.contains("DifferentIndividuals")) ||
                  		 (word.contains("DisjointUnion")))) {
                     if(!(uniqueClasses.contains(word))) {
                        uniqueClasses.add(word);
                     }
                  }
               	
               }
            }
         }
      } 
      catch (FileNotFoundException e) {
         e.printStackTrace();
      } 
      catch (IOException e) {
         e.printStackTrace();
      }
      return new HashSet<String>(uniqueClasses);
   }
	
	// Read in translations from file
	//Break words up on capital letters and try to translate them individually
	//Write translated words to file to be used.
   public static ArrayList<String[]> translateClasses(Set<String> wordsForTranslation) {
      wordsForTranslation = new HashSet<String>(wordsForTranslation);
   
   	//Scanner is used here because buffered readers can't read big files.
      ArrayList<String[]> outArr = new ArrayList<String[]>();
      Scanner sc = null;
      try {
      	//Read in from the afrikaans dictionary file. Its simply a text file.
      	//The file is of the form:  'afrikaansWord = englishWord'
         sc = new Scanner(new File("translation/AfrikaansToEnglish.dict"));
      } 
      catch (FileNotFoundException e) {
         e.printStackTrace();
      }
      String line; 
         //iterate over file and add words which can be translated to the output array.
      while (sc.hasNextLine()) {
         line = sc.nextLine();
         String[] lineArr = line.split("=");
         String afrikaansWord = lineArr[0];
         String englishWord = lineArr[1].trim();
         englishWord = capitalise(englishWord);
         if (wordsForTranslation.contains(englishWord)) {
            wordsForTranslation.remove(englishWord);
            outArr.add(new String[]{englishWord, afrikaansWord});
         }
      }
      sc.close();
      return outArr;
   }

	//normalises the word so that it can be compared, deals with certain issues such as the removal of dashes, spaces and capitalisation.
   private static String capitalise(String str) {
   	//Only capitalise if this is lowercase
      if(!(str.toLowerCase()).equals(str)){
         return str;
      }
      String outS="";
      String[] arr = null;
   	
      if (str.lastIndexOf(" ") > 0) {
         arr = str.split(" ");
      }
      else if (str.contains("-")) {
         arr = str.split("-");
      }
      else {
         return str.substring(0, 1).toUpperCase() + str.substring(1).toLowerCase();
      }
      for(String word : arr) {
         outS+= word.substring(0, 1).toUpperCase() + word.substring(1);
      }
      return outS;
   }
	//Starts the process of creating GF files by calling the nested Absctract and concrete syntax methods
   public static void createGFFiles(Set<String> wordsForTranslation, ArrayList<String[]> translatedWords) throws FileNotFoundException {
      createAbstractSyntax(wordsForTranslation);
      createConcreteSyntax(translatedWords);
   }
	
   public static void createConcreteSyntax(ArrayList<String[]> translatedWords) throws FileNotFoundException {
      String prependText = "--# -path=present\n\n\nconcrete OwlAfr of Owl = open LangAfr in {\nflags coding=utf8;\n\nlincat\n" + 
         "Ind = {s : Str};\nProp = {s:Str};\nlin";
      String appendText = "}";
      String wordsFile = gf_path+"\\lib\\own\\OwlAfr.gf";
      String axiomsFile = "translation/OwlAxioms.txt";
      PrintWriter writer = new PrintWriter(wordsFile);
      writer.println(prependText);
      for (int i = 0; i < translatedWords.size(); i++) {
         String[] arr =  (String[]) translatedWords.get(i);
         writer.println(arr[0] + "= {s=\"" + arr[1].trim()+"\"};");
      }
      try (BufferedReader br = new BufferedReader(new FileReader(axiomsFile))) {
         String line;
         while ((line = br.readLine()) != null) {
            writer.println(line);
         }
      } 
      catch (FileNotFoundException e) {
         System.out.println("ERROR! Unable to read file " +axiomsFile+". Please check if it exists.");
         e.printStackTrace();
      } 
      catch (IOException e) {
      		// TODO Auto-generated catch block
         System.out.println("ERROR! I/O Exception while reading file.");
         e.printStackTrace();
      }
   		
      writer.println(appendText);
      writer.close();
   }
   public static void createAbstractSyntax(Set<String> wordsForTranslation) {
   	//Add file with all the words we need.
   	//Hardcoded URL for now
      String prependText = "abstract OntologySpecificClasses = Cat **\n{\ncat\nInd; Prop;\n\nfun";
      String appendText = " : Prop;";
      String filename = "translation/AfrikaansTranslation.txt";
      String url = gf_path+"\\lib\\own\\OntologySpecificClasses.gf";
      PrintWriter classWriter = null;
   
      try {
         classWriter = new PrintWriter(url);
      } 
      catch (FileNotFoundException e) {
         System.out.println("ERROR! File not found at the following path: "+ url);
         e.printStackTrace();
      }
      classWriter.println(prependText);
   		
      for(String word : wordsForTranslation) {
         classWriter.println(word+appendText);
      }
   		
   		
   	//Close off the class with a closing bracket.
      classWriter.println("}");
      classWriter.close();
   }
	//Writes Axioms from functionalSyntax to a format which GF accepts/excepts.
   public static void prepareForLinearisation(String ontologyName) throws FileNotFoundException {
      String writeFilename = "translation/axioms.gfs";
      String prependText = "l ";
      String appendText = " | write_file -append -file=webpage/" + ontologyName +".txt";
      String filename = "functionalSyntax.txt";
      PrintWriter axiomwriter = new PrintWriter(writeFilename);
      axiomwriter.println("i ..\\own\\Owl.gf ..\\own\\OwlAfr.gf");
      try (BufferedReader br = new BufferedReader(new FileReader(filename))) {
         String line;
         while ((line = br.readLine()) != null) {
            axiomwriter.println(prependText +line+ appendText);
         }
      } 
      catch (FileNotFoundException e) {
         System.out.println("ERROR! Unable to read file " +filename+". Please check if it exists.");
         e.printStackTrace();
      } 
      catch (IOException e) {
      		// TODO Auto-generated catch block
         System.out.println("ERROR! I/O Exception while reading file.");
         e.printStackTrace();
      }
      axiomwriter.close(); 
   }	
	//Courtesy of User CraigTP on StackOverflow
	//Credit: http://stackoverflow.com/questions/1102891/how-to-check-if-a-string-is-a-numeric-type-in-java
   private static boolean isNumeric(String str)  
   {  
      try  
      {  
         double d = Double.parseDouble(str);  
      }  
      catch(NumberFormatException nfe)  
      {  
         return false;  
      }  
      return true;  
   }
}
